# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy
import hashlib
from unittest import mock

import six
import uuid

from osc_lib.tests import utils

from qinlingclient.tests.unit import fakes


class FakeQinlingClient(object):

    def __init__(self, **kwargs):
        self.auth_token = kwargs['auth_token']
        self.auth_url = kwargs['auth_url']
        self.runtimes = mock.Mock()
        self.functions = mock.Mock()
        self.function_executions = mock.Mock()
        self.function_versions = mock.Mock()
        self.function_workers = mock.Mock()
        self.jobs = mock.Mock()
        self.webhooks = mock.Mock()
        self.function_aliases = mock.Mock()


class TestQinlingClient(utils.TestCommand):

    def setUp(self):
        super(TestQinlingClient, self).setUp()
        self.app.client_manager.function_engine = FakeQinlingClient(
            auth_token=fakes.AUTH_TOKEN,
            auth_url=fakes.AUTH_URL
        )


class FakeRuntime(object):
    """Fake one or more runtimes."""

    @staticmethod
    def create_one_runtime(attrs=None):
        """Create a fake runtime.

        :param Dictionary attrs:
            A dictionary with all atrributes
        :return:
            A FakeResource object, with id, name, etc.
        """

        attrs = attrs or {}
        # Set default attributes.
        runtime_attrs = {
            'id': str(uuid.uuid4()),
            'name': 'runtime-name-' + uuid.uuid4().hex,
            'image': 'openstackqinling/python-runtime',
            'status': 'available',
            'description': 'runtime-description-' + uuid.uuid4().hex,
            'is_public': True,
            'trusted': True,
            'project_id': str(uuid.uuid4()),
            'created_at': '2018-07-26 09:00:00',
            'updated_at': '2018-07-26 09:00:30'
        }

        # Overwrite default attributes.
        runtime_attrs.update(attrs)

        runtime = fakes.FakeResource(info=copy.deepcopy(runtime_attrs),
                                     loaded=True)

        return runtime

    @staticmethod
    def create_runtimes(attrs=None, count=2):
        """Create multiple fake runtimes.

        :param Dictionary attrs:
            A dictionary with all atrributes
        :param int count:
            The number of runtimes to fake
        :return:
            A list of FakeResource objects faking the runtimes.
        """

        runtimes = []
        for i in range(count):
            runtimes.append(FakeRuntime.create_one_runtime(attrs))

        return runtimes

    @staticmethod
    def get_runtimes(runtimes=None, count=2):
        """Get an iterable Mock object with a list of faked runtimes.

        If runtimes list is provided, then initialize the Mock object with the
        list. Otherwise create one.

        :param List runtimes:
            A list of FakeResource faking runtimes
        :param int count:
            The number of runtimes to fake
        :return:
            An iterable Mock object with side_effect set to a list of faked
            runtimes.
        """

        if runtimes is None:
            runtimes = FakeRuntime.create_runtimes(count=count)

        return mock.Mock(side_effect=runtimes)

    @staticmethod
    def create_one_runtime_pool(attrs=None):
        """Create a fake runtime pool.

        :param Dictionary attrs:
            A dictionary with all atrributes
        :return:
            A FakeResource object, with name, capacity.
        """

        attrs = attrs or {}
        # Set default attributes.
        pool_attrs = {
            'name': 'runtime-id-' + uuid.uuid4().hex,
            'capacity': {'available': 5, 'total': 5}
        }

        # Overwrite default attributes.
        pool_attrs.update(attrs)

        runtime_pool = fakes.FakeResource(info=copy.deepcopy(pool_attrs),
                                          loaded=True)

        return runtime_pool


class FakeFunction(object):
    """Fake one or more functions."""

    @staticmethod
    def get_fake_md5():
        content = uuid.uuid4().hex
        if not isinstance(content, six.binary_type):
            content = content.encode('utf-8')
        return hashlib.md5(content).hexdigest()

    @staticmethod
    def create_one_function(attrs=None):
        """Create a fake function.

        :param Dictionary attrs:
            A dictionary with all atrributes
        :return:
            A FakeResource object, with id, name, etc.
        """

        attrs = attrs or {}
        # Set default attributes.
        function_attrs = {
            'id': str(uuid.uuid4()),
            'name': 'function-name-' + uuid.uuid4().hex,
            'description': 'function-description-' + uuid.uuid4().hex,
            'count': 0,
            'code': {'md5sum': FakeFunction.get_fake_md5(),
                     'source': 'package'},
            'runtime_id': str(uuid.uuid4()),
            'entry': 'main.main',
            'project_id': str(uuid.uuid4()),
            'created_at': '2018-07-26 09:00:00',
            'updated_at': '2018-07-26 09:00:30',
            'cpu': 100,
            'memory_size': 33554432,
            'timeout': 5
        }

        # Overwrite default attributes.
        function_attrs.update(attrs)

        function = fakes.FakeResource(info=copy.deepcopy(function_attrs),
                                      loaded=True)

        return function

    @staticmethod
    def create_functions(attrs=None, count=2):
        """Create multiple fake functions.

        :param Dictionary attrs:
            A dictionary with all atrributes
        :param int count:
            The number of functions to fake
        :return:
            A list of FakeResource objects faking the functions.
        """

        functions = []
        for i in range(count):
            functions.append(FakeFunction.create_one_function(attrs))

        return functions

    @staticmethod
    def get_functions(functions=None, count=2):
        """Get an iterable Mock object with a list of faked functions.

        If functions list is provided, then initialize the Mock object with the
        list. Otherwise create one.

        :param List functions:
            A list of FakeResource faking functions
        :param int count:
            The number of functions to fake
        :return:
            An iterable Mock object with side_effect set to a list of faked
            functions.
        """

        if functions is None:
            functions = FakeFunction.create_functions(count=count)

        return mock.Mock(side_effect=functions)


class FakeExecution(object):
    """Fake one or more function executions."""

    @staticmethod
    def create_one_execution(attrs=None):
        """Create a fake function execution.

        :param Dictionary attrs:
            A dictionary with all atrributes
        :return:
            A FakeResource object, with id, function_id, etc.
        """

        attrs = attrs or {}
        # Set default attributes.
        execution_attrs = {
            'id': str(uuid.uuid4()),
            'function_alias': None,
            'function_id': str(uuid.uuid4()),
            'function_version': 0,
            'description': 'execution-description-' + uuid.uuid4().hex,
            'input': '{"FAKE_INPUT_KEY": "FAKE_INPUT_VALUE"}',
            'result': '{"duration": 1.234, "output": "FAKE_OUTPUT"}',
            'status': 'success',
            'sync': True,
            'project_id': str(uuid.uuid4()),
            'created_at': '2018-07-26 09:00:00',
            'updated_at': '2018-07-26 09:00:30',
        }

        # Overwrite default attributes.
        execution_attrs.update(attrs)

        execution = fakes.FakeResource(info=copy.deepcopy(execution_attrs),
                                       loaded=True)

        return execution

    @staticmethod
    def create_executions(attrs=None, count=2):
        """Create multiple fake function executions.

        :param Dictionary attrs:
            A dictionary with all atrributes
        :param int count:
            The number of function executions to fake
        :return:
            A list of FakeResource objects faking the function executions.
        """

        executions = []
        for i in range(count):
            executions.append(FakeExecution.create_one_execution(attrs))

        return executions

    @staticmethod
    def get_executions(executions=None, count=2):
        """Get an iterable Mock object with a list of faked executions.

        If function executions list is provided, then initialize the Mock
        object with the list. Otherwise create one.

        :param List executions:
            A list of FakeResource faking function executions
        :param int count:
            The number of function executions to fake
        :return:
            An iterable Mock object with side_effect set to a list of faked
            function executions.
        """

        if executions is None:
            executions = FakeExecution.create_executions(count=count)

        return mock.Mock(side_effect=executions)


class FakeFunctionVersion(object):
    """Fake one or more function versions."""

    @staticmethod
    def create_one_function_version(attrs=None):
        """Create a fake function version.

        :param Dictionary attrs:
            A dictionary with all atrributes
        :return:
            A FakeResource object, with id, function_id, etc.
        """

        attrs = attrs or {}
        # Set default attributes.
        function_version_attrs = {
            'id': str(uuid.uuid4()),
            'function_id': str(uuid.uuid4()),
            'description': 'function-version-description-' + uuid.uuid4().hex,
            'version_number': 1,
            'count': 0,
            'project_id': str(uuid.uuid4()),
            'created_at': '2018-07-26 09:00:00',
            'updated_at': '2018-07-26 09:00:30',
        }

        # Overwrite default attributes.
        function_version_attrs.update(attrs)

        function_version = fakes.FakeResource(
            info=copy.deepcopy(function_version_attrs),
            loaded=True)

        return function_version

    @staticmethod
    def create_function_versions(attrs=None, count=2):
        """Create multiple fake function versions.

        :param Dictionary attrs:
            A dictionary with all atrributes
        :param int count:
            The number of function versions to fake
        :return:
            A list of FakeResource objects faking the function versions.
        """

        function_versions = []
        for i in range(count):
            function_versions.append(
                FakeFunctionVersion.create_one_function_version(attrs)
            )

        return function_versions

    @staticmethod
    def get_function_versions(function_versions=None, count=2):
        """Get an iterable Mock object with a list of faked function versions.

        If function versions list is provided, then initialize the Mock
        object with the list. Otherwise create one.

        :param List function_versions:
            A list of FakeResource faking function versions
        :param int count:
            The number of function versions to fake
        :return:
            An iterable Mock object with side_effect set to a list of faked
            function versions.
        """

        if function_versions is None:
            function_versions = FakeFunctionVersion.create_function_versions(
                count=count
            )

        return mock.Mock(side_effect=function_versions)


class FakeFunctionWorker(object):
    """Fake one or more function workers."""

    @staticmethod
    def create_one_function_worker(attrs=None):
        """Create a fake function worker.

        :param Dictionary attrs:
            A dictionary with all atrributes
        :return:
            A FakeResource object, with function_id, worker_name, etc.
        """

        attrs = attrs or {}
        # Set default attributes.
        function_worker_attrs = {
            'function_id': str(uuid.uuid4()),
            'worker_name': 'worker-' + uuid.uuid4().hex,
        }

        # Overwrite default attributes.
        function_worker_attrs.update(attrs)

        function_worker = fakes.FakeResource(
            info=copy.deepcopy(function_worker_attrs),
            loaded=True)

        return function_worker

    @staticmethod
    def create_function_workers(attrs=None, count=2):
        """Create multiple fake function workers.

        :param Dictionary attrs:
            A dictionary with all atrributes
        :param int count:
            The number of function workers to fake
        :return:
            A list of FakeResource objects faking the function workers.
        """

        function_workers = []
        for i in range(count):
            function_workers.append(
                FakeFunctionWorker.create_one_function_worker(attrs)
            )

        return function_workers

    @staticmethod
    def get_function_workers(function_workers=None, count=2):
        """Get an iterable Mock object with a list of faked function workers.

        If function workers list is provided, then initialize the Mock
        object with the list. Otherwise create one.

        :param List function_workers:
            A list of FakeResource faking function workers
        :param int count:
            The number of function workers to fake
        :return:
            An iterable Mock object with side_effect set to a list of faked
            function workers.
        """

        if function_workers is None:
            function_workers = FakeFunctionWorker.create_function_workers(
                count=count
            )

        return mock.Mock(side_effect=function_workers)


class FakeJob(object):
    """Fake one or more jobs."""

    @staticmethod
    def create_one_job(attrs=None):
        """Create a fake job.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with id, name, etc.
        """

        attrs = attrs or {}
        # Set default attributes.
        job_attrs = {
            'id': str(uuid.uuid4()),
            'name': 'job-name-' + uuid.uuid4().hex,
            'count': 3,
            'status': 'RUNNING',
            'function_alias': None,
            'function_id': str(uuid.uuid4()),
            'function_version': 0,
            'function_input': '{"FAKE_INPUT_KEY": "FAKE_INPUT_VALUE"}',
            'pattern': '0 * * * *',  # Once per hour
            'first_execution_time': '2018-08-08T08:00:00',
            'next_execution_time': '2018-08-08T10:00:00',
            'project_id': str(uuid.uuid4()),
            'created_at': '2018-07-26 09:00:00',
            'updated_at': '2018-07-26 09:00:30',
        }

        # Overwrite default attributes.
        job_attrs.update(attrs)

        job = fakes.FakeResource(info=copy.deepcopy(job_attrs), loaded=True)
        return job

    @staticmethod
    def create_jobs(attrs=None, count=2):
        """Create multiple fake jobs.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of jobs to fake
        :return:
            A list of FakeResource objects faking the jobs.
        """

        jobs = []
        for i in range(count):
            jobs.append(FakeJob.create_one_job(attrs))

        return jobs

    @staticmethod
    def get_jobs(jobs=None, count=2):
        """Get an iterable mock object with a list of faked jobs.

        If jobs list is provided, then initialize the Mock object with the
        list. Otherwise create one.

        :param List jobs:
            A list of FakeResource faking jobs
        :param int count:
            The number of jobs to fake
        :return:
            An iterable Mock object with side_effect set to a list of faked
            jobs.
        """

        if jobs is None:
            jobs = FakeJob.create_jobs(count=count)

        return mock.Mock(side_effect=jobs)


class FakeWebhook(object):
    """Fake one or more webhooks."""

    @staticmethod
    def create_one_webhook(attrs=None):
        """Create a fake webhook.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with id, function_id, etc.
        """

        attrs = attrs or {}
        # Set default attributes.
        webhook_attrs = {
            'id': str(uuid.uuid4()),
            'function_alias': None,
            'function_id': str(uuid.uuid4()),
            'function_version': 0,
            'description': 'webhook-description-' + uuid.uuid4().hex,
            'project_id': str(uuid.uuid4()),
            'created_at': '2018-07-26 09:00:00',
            'updated_at': '2018-07-26 09:00:30',
            'webhook_url': 'http://HOST:PORT/v1/webhooks/FAKE_ID/invoke',
        }

        # Overwrite default attributes.
        webhook_attrs.update(attrs)

        webhook = fakes.FakeResource(info=copy.deepcopy(webhook_attrs),
                                     loaded=True)
        return webhook

    @staticmethod
    def create_webhooks(attrs=None, count=2):
        """Create multiple fake webhooks.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of webhooks to fake
        :return:
            A list of FakeResource objects faking the webhooks.
        """

        webhooks = []
        for i in range(count):
            webhooks.append(FakeWebhook.create_one_webhook(attrs))

        return webhooks

    @staticmethod
    def get_webhooks(webhooks=None, count=2):
        """Get an iterable mock object with a list of faked webhooks.

        If webhooks list is provided, then initialize the Mock object with the
        list. Otherwise create one.

        :param List webhooks:
            A list of FakeResource faking webhooks
        :param int count:
            The number of webhooks to fake
        :return:
            An iterable Mock object with side_effect set to a list of faked
            webhooks.
        """

        if webhooks is None:
            webhooks = FakeWebhook.create_webhooks(count=count)

        return mock.Mock(side_effect=webhooks)


class FakeFunctionAlias(object):
    """Fake one or more function aliases."""

    @staticmethod
    def create_one_function_alias(attrs=None):
        """Create a fake function alias.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A FakeResource object, with name, function_id, etc.
        """

        attrs = attrs or {}
        # Set default attributes
        function_alias_attrs = {
            'name': 'function-alias-name-' + uuid.uuid4().hex,
            'function_id': str(uuid.uuid4()),
            'description': 'function-alias-description-' + uuid.uuid4().hex,
            'function_version': 0,
            'project_id': str(uuid.uuid4()),
            'created_at': '2018-07-26 09:00:00',
            'updated_at': '2018-07-26 09:00:30',
        }

        # Overwrite default attributes
        function_alias_attrs.update(attrs)

        function_alias = fakes.FakeResource(
            info=copy.deepcopy(function_alias_attrs), loaded=True)
        return function_alias

    @staticmethod
    def create_function_aliases(attrs=None, count=2):
        """Create multiple fake function aliases.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of function aliases to fake
        :return:
            A list of FakeResource objects faking the function aliases.
        """

        function_aliases = []
        for i in range(count):
            function_aliases.append(
                FakeFunctionAlias.create_one_function_alias(attrs)
            )

        return function_aliases

    @staticmethod
    def get_function_aliases(function_aliases=None, count=2):
        """Get an iterable mock object with a list of faked function aliases.

        If function aliases list is provided, then initialize the Mock object
        with the list. Otherwise create one.

        :param List function_aliases:
            A list of FakeResource faking function aliases
        :param int count:
            The number of function aliases to fake
        :return
            An iterable Mock object with side_effect set to a list of faked
            function aliases.
        """

        if function_aliases is None:
            function_aliases = FakeFunctionAlias.create_function_aliases(
                count=count
            )

        return mock.Mock(side_effect=function_aliases)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_serialization import jsonutils

from qinlingclient.common import base


class Function(base.Resource):
    pass


class FunctionManager(base.ManagerWithFind):
    resource_class = Function

    def list(self, **kwargs):
        q_list = []
        for key, value in kwargs.items():
            q_list.append('%s=%s' % (key, value))
        q_params = '&'.join(q_list)

        url = '/v1/functions'
        if q_params:
            url += '?%s' % q_params
        return self._list(url, response_key='functions')

    def create(self, code, runtime=None, package=None, **kwargs):
        data = {
            'runtime_id': runtime,
            'code': jsonutils.dumps(code)
        }

        for k, v in kwargs.items():
            if v is not None:
                data.update({k: v})

        params = {"data": data}
        if package:
            params.update({"files": {'package': package}})

        response = self.http_client.request(
            '/v1/functions',
            'POST',
            **params
        )
        body = jsonutils.loads(response.text)

        return self.resource_class(self, body)

    def delete(self, id):
        self._delete('/v1/functions/%s' % id)

    def get(self, id, download=False):
        url = '/v1/functions/%s' % id
        if not download:
            return self._get('/v1/functions/%s' % id)

        url = url + '?download=true'
        return self.http_client.request(url, 'GET', stream=True)

    def update(self, id, code=None, package=None, **kwargs):
        if code:
            kwargs.update(code)

        params = {"data": kwargs}
        if package:
            params.update({"files": {'package': package}})

        response = self.http_client.request(
            '/v1/functions/%s' % id,
            'PUT',
            **params
        )
        body = jsonutils.loads(response.text)

        return self.resource_class(self, body)

    def detach(self, id):
        return self.http_client.request(
            '/v1/functions/%s/detach' % id,
            'POST',
        )

    def scaleup(self, id, count=1):
        params = {'data': {'count': count}}
        return self.http_client.json_request(
            '/v1/functions/%s/scale_up' % id,
            'POST',
            **params
        )

    def scaledown(self, id, count=1):
        params = {'data': {'count': count}}
        return self.http_client.json_request(
            '/v1/functions/%s/scale_down' % id,
            'POST',
            **params
        )



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import uuid

from qinlingclient.tests.unit.v1 import test_client

WORKER_1 = {'function_id': str(uuid.uuid4()), 'worker_name': 'worker_1'}
WORKER_2 = {'function_id': WORKER_1['function_id'], 'worker_name': 'worker_2'}

LIST_FUNCTION_WORKERS_RESP = {
    'workers': [WORKER_1, WORKER_2]
}


class TestFunctionWorker(test_client.TestQinlingClient):

    def test_list_function_worker(self):
        function_id = WORKER_1['function_id']
        self.requests_mock.register_uri(
            'GET',
            test_client.QINLING_URL + '/v1/functions/%s/workers' % function_id,
            headers={'Content-Type': 'application/json'},
            json=LIST_FUNCTION_WORKERS_RESP,
            status_code=200
        )
        ret = self.client.function_workers.list(function_id)
        self.assertIsInstance(ret, list)
        self.assertEqual(
            [WORKER_1, WORKER_2],
            [resource.to_dict() for resource in ret])



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from qinlingclient.common import base


class Webhook(base.Resource):
    pass


class WebhookManager(base.Manager):
    resource_class = Webhook

    def list(self, **kwargs):
        q_list = []
        for key, value in kwargs.items():
            q_list.append('%s=%s' % (key, value))
        q_params = '&'.join(q_list)

        url = '/v1/webhooks'
        if q_params:
            url += '?%s' % q_params
        return self._list(url, response_key='webhooks')

    def create(self, function_id, function_alias=None, function_version=0,
               description=None):
        data = {
            'function_id': function_id,
            'function_version': function_version,
            'function_alias': function_alias
        }
        if description:
            data.update({'description': description})

        return self._create('/v1/webhooks', data)

    def delete(self, id):
        self._delete('/v1/webhooks/%s' % id)

    def get(self, id):
        return self._get('/v1/webhooks/%s' % id)

    def update(self, id, **kwargs):
        """Update webhook.

        function_id, function_version and description are supported.
        """
        body = {}
        for k, v in kwargs.items():
            if v is not None:
                body.update({k: v})

        return self._update('/v1/webhooks/%s' % id, kwargs)



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from osc_lib.tests import utils as osc_tests_utils

from qinlingclient.common import exceptions
from qinlingclient.osc.v1 import base
from qinlingclient.osc.v1 import webhook
from qinlingclient.tests.unit.osc.v1 import fakes


class TestWebhook(fakes.TestQinlingClient):

    def setUp(self):
        super(TestWebhook, self).setUp()
        # Get a shortcut
        self.client = self.app.client_manager.function_engine

        self.columns = base.WEBHOOK_COLUMNS
        self.data = []

        webhooks = fakes.FakeWebhook.create_webhooks(count=3)
        self._webhooks = webhooks
        for w in self._webhooks:
            self.data.append((w.id, w.function_alias, w.function_id,
                              w.function_version, w.description, w.project_id,
                              w.created_at, w.updated_at,
                              w.webhook_url))


class TestListWebhook(TestWebhook):

    def setUp(self):
        super(TestListWebhook, self).setUp()
        self.cmd = webhook.List(self.app, None)

        self.columns = [c.capitalize() for c in base.WEBHOOK_COLUMNS]

        self.client.webhooks.list = mock.Mock(return_value=self._webhooks)

    def test_webhook_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.client.webhooks.list.assert_called_once_with()
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_webhook_list_with_filter(self):
        arglist = ['--filter', 'function_version=neq:0',
                   '--filter', 'description=has:webhook']
        verifylist = [
            ('filters', ['function_version=neq:0', 'description=has:webhook']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.client.webhooks.list.assert_called_once_with(
            function_version='neq:0', description='has:webhook'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_webhook_list_with_invalid_filter(self):
        arglist = ['--filter', 'function_version']
        verifylist = [
            ('filters', ['function_version']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            ValueError,
            '^Invalid filter: function_version$',
            self.cmd.take_action, parsed_args
        )


class TestCreateWebhook(TestWebhook):

    def setUp(self):
        super(TestCreateWebhook, self).setUp()
        self.cmd = webhook.Create(self.app, None)

    def _create_fake_webhook(self, attrs=None):
        # Allow to fake different create results
        w = fakes.FakeWebhook.create_one_webhook(attrs)
        self.client.webhooks.create = mock.Mock(return_value=w)
        data = (w.id, w.function_alias, w.function_id, w.function_version,
                w.description, w.project_id, w.created_at, w.updated_at,
                w.webhook_url)
        return data

    def test_webhook_create_function_id(self):
        """Create a webhook with function id."""
        function_id = self._webhooks[0].function_id
        attrs = {'function_id': function_id}
        created_data = self._create_fake_webhook(attrs)

        arglist = ['--function', function_id]
        verifylist = [
            ('function', function_id),
            ('function_version', 0),
            ('function_alias', None),
            ('description', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.webhooks.create.assert_called_once_with(
            **{'function_id': function_id,
               'function_version': 0,
               'function_alias': None,
               'description': None}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

    def test_webhook_create_function_name(self):
        """Create a webhook.

        1. use function name to find the function_id,
        2. all optional params are specified.
        """
        function = fakes.FakeFunction.create_one_function()
        function_name = function.name
        function_id = function.id
        function_version = 1
        webhook_description = 'This is a newly created webhook.'
        attrs = {'function_id': function_id,
                 'function_version': function_version,
                 'description': webhook_description}
        created_data = self._create_fake_webhook(attrs)

        # Use to find the function id with its name
        self.client.functions.find.return_value = function

        arglist = ['--function', function_name,
                   '--function-version', str(function_version),
                   '--description', webhook_description]
        verifylist = [
            ('function', function_name),
            ('function_version', function_version),
            ('function_alias', None),
            ('description', webhook_description),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.webhooks.create.assert_called_once_with(
            **{'function_id': function_id,
               'function_version': function_version,
               'function_alias': None,
               'description': webhook_description}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

        self.client.functions.find.assert_called_once_with(name=function_name)

    def test_webhook_create_function_alias(self):
        """Create a webhook with function alias."""
        function_alias = 'fake_alias'
        attrs = {'function_alias': function_alias}
        created_data = self._create_fake_webhook(attrs)

        arglist = ['--function-alias', function_alias]
        verifylist = [
            ('function', None),
            ('function_version', 0),
            ('function_alias', function_alias),
            ('description', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.webhooks.create.assert_called_once_with(
            **{'function_id': None,
               'function_version': None,
               'function_alias': function_alias,
               'description': None}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

    def test_webhook_create_version_not_integer(self):
        # function_version should be an integer value
        function_id = self._webhooks[0].function_id

        arglist = [function_id, '--function-version', 'NOT_A_INTEGER']
        verifylist = [
            ('function', function_id),
            ('function_version', 'NOT_A_INTEGER'),
            ('description', None),
        ]

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestDeleteWebhook(TestWebhook):

    def setUp(self):
        super(TestDeleteWebhook, self).setUp()
        self.cmd = webhook.Delete(self.app, None)
        self.client.webhooks.delete = mock.Mock(return_value=None)

    def test_webhook_delete_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_webhook_delete_one(self):
        webhook_id = self._webhooks[0].id
        arglist = [webhook_id]
        verifylist = [('webhook', [webhook_id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.webhooks.delete.assert_called_once_with(webhook_id)

    def test_webhook_delete_multiple(self):
        webhook_ids = [w.id for w in self._webhooks]
        arglist = webhook_ids
        verifylist = [('webhook', webhook_ids)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        calls = [mock.call(w_id) for w_id in webhook_ids]
        self.assertEqual(len(webhook_ids),
                         self.client.webhooks.delete.call_count)
        self.client.webhooks.delete.assert_has_calls(calls)

    def test_webhook_delete_multiple_exception(self):
        webhook_ids = [w.id for w in self._webhooks]
        arglist = webhook_ids
        verifylist = [('webhook', webhook_ids)]

        self.client.webhooks.delete = mock.Mock(side_effect=[
            None, RuntimeError, None
        ])

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Unable to delete the specified webhook\(s\)\.$',
            self.cmd.take_action, parsed_args)

        # The second deleteion failed, but the third is done normally
        calls = [mock.call(w_id) for w_id in webhook_ids]
        self.assertEqual(len(webhook_ids),
                         self.client.webhooks.delete.call_count)
        self.client.webhooks.delete.assert_has_calls(calls)


class TestShowWebhook(TestWebhook):

    def setUp(self):
        super(TestShowWebhook, self).setUp()
        self.cmd = webhook.Show(self.app, None)
        self.client.webhooks.get = mock.Mock(return_value=self._webhooks[0])

    def test_webhook_show_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_webhook_show(self):
        webhook_id = self._webhooks[0].id
        arglist = [webhook_id]
        verifylist = [('webhook', webhook_id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.webhooks.get.assert_called_once_with(webhook_id)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data[0], data)


class TestUpdateWebhook(TestWebhook):

    def setUp(self):
        super(TestUpdateWebhook, self).setUp()
        self.cmd = webhook.Update(self.app, None)

    def _update_fake_webhook(self, attrs=None):
        # Allow to fake different update results
        w = fakes.FakeWebhook.create_one_webhook(attrs)
        self.client.webhooks.update = mock.Mock(return_value=w)
        data = (w.id, w.function_alias, w.function_id, w.function_version,
                w.description, w.project_id, w.created_at, w.updated_at,
                w.webhook_url)
        return data

    def test_webhook_update_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_webhook_update_required_options(self):
        """Update a webhook.

        Do nothing as only the webhook id is specified.
        """
        webhook_id = self._webhooks[0].id
        attrs = {'id': webhook_id}
        updated_data = self._update_fake_webhook(attrs)

        arglist = [webhook_id]
        verifylist = [
            ('id', webhook_id),
            ('function_id', None),
            ('function_version', None),
            ('description', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.webhooks.update.assert_called_once_with(
            webhook_id,
            **{'function_id': None,
               'function_version': None,
               'description': None}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(updated_data, data)

    def test_webhook_update_all_options(self):
        webhook_id = self._webhooks[0].id
        function_id = self._webhooks[1].function_id
        function_version = 1
        webhook_description = 'This is a updated webhook.'
        attrs = {'id': webhook_id, 'function_id': function_id,
                 'function_version': function_version,
                 'description': webhook_description}
        updated_data = self._update_fake_webhook(attrs)

        arglist = [webhook_id, '--function-id', function_id,
                   '--function-version', str(function_version),
                   '--description', webhook_description]
        verifylist = [
            ('id', webhook_id),
            ('function_id', function_id),
            ('function_version', str(function_version)),
            ('description', webhook_description),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.webhooks.update.assert_called_once_with(
            webhook_id,
            **{'function_id': function_id,
               'function_version': str(function_version),
               'description': webhook_description}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(updated_data, data)

    def test_webhook_update_version_not_integer(self):
        # function_version should be an integer value
        webhook_id = self._webhooks[0].id
        function_id = self._webhooks[0].function_id

        arglist = [webhook_id, function_id,
                   '--function-version', 'NOT_A_INTEGER']
        verifylist = [
            ('id', webhook_id),
            ('function_id', function_id),
            ('function_version', 'NOT_A_INTEGER'),
            ('description', None),
        ]

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)



# Copyright 2018 OpenStack Foundation.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from osc_lib.command import command
from osc_lib import utils
from oslo_utils import uuidutils

from qinlingclient.osc.v1 import base
from qinlingclient import utils as q_utils


class List(base.QinlingLister):
    columns = base.FUNCTION_ALIAS_COLUMNS
    filtered_columns = base.FILTERED_FUNCTION_ALIAS_COLUMNS

    def _get_resources(self, parsed_args):
        client = self.app.client_manager.function_engine

        return client.function_aliases.list(**base.get_filters(parsed_args))


class Create(command.ShowOne):
    columns = base.FUNCTION_ALIAS_COLUMNS

    def get_parser(self, prog_name):
        parser = super(Create, self).get_parser(prog_name)

        parser.add_argument(
            "name",
            help="Function Alias name.",
        )
        parser.add_argument(
            "--function",
            required=True,
            help="Function ID or Name.",
        )
        parser.add_argument(
            "--function-version",
            type=int,
            default=0,
            help="Function Version number.",
        )
        parser.add_argument(
            "--description",
            default='',
            help="Description for the new alias.",
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine

        function_id = parsed_args.function
        if not uuidutils.is_uuid_like(function_id):
            # Try to find the function id with name
            function_id = q_utils.find_resource_id_by_name(
                client.functions, function_id)

        alias = client.function_aliases.create(
            parsed_args.name,
            function_id=function_id,
            function_version=parsed_args.function_version,
            description=parsed_args.description,
        )

        return self.columns, utils.get_item_properties(alias, self.columns)


class Delete(base.QinlingDeleter):
    def get_parser(self, prog_name):
        parser = super(Delete, self).get_parser(prog_name)

        parser.add_argument(
            "name",
            nargs='+',
            help="Function Alias name(s).",
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine

        self.delete = client.function_aliases.delete
        self.resource = 'function_alias'

        self.delete_resources(parsed_args.name)


class Show(command.ShowOne):
    columns = base.FUNCTION_ALIAS_COLUMNS

    def get_parser(self, prog_name):
        parser = super(Show, self).get_parser(prog_name)

        parser.add_argument(
            "name",
            help="Function Alias name.",
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine

        alias = client.function_aliases.get(parsed_args.name)

        return self.columns, utils.get_item_properties(alias, self.columns)


class Update(command.ShowOne):
    columns = base.FUNCTION_ALIAS_COLUMNS

    def get_parser(self, prog_name):
        parser = super(Update, self).get_parser(prog_name)

        parser.add_argument(
            "name",
            help="Function Alias name.",
        )
        parser.add_argument(
            "--function",
            help="Function ID or Name.",
        )
        parser.add_argument(
            "--function-version",
            help="Function Version number.",
        )
        parser.add_argument(
            "--description",
            help="Description for the new alias.",
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine

        function_id = parsed_args.function
        if function_id and not uuidutils.is_uuid_like(function_id):
            # Try to find the function id with name
            function_id = q_utils.find_resource_id_by_name(
                client.functions, function_id)

        alias = client.function_aliases.update(
            parsed_args.name,
            function_id=function_id,
            function_version=parsed_args.function_version,
            description=parsed_args.description,
        )

        return self.columns, utils.get_item_properties(alias, self.columns)



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock
import zipfile

from osc_lib.tests import utils as osc_tests_utils
import testtools

from qinlingclient.common import exceptions
from qinlingclient.osc.v1 import base
from qinlingclient.osc.v1 import function
from qinlingclient.tests.unit.osc.v1 import fakes


class TestFunctionUtils(testtools.TestCase):

    def setUp(self):
        super(TestFunctionUtils, self).setUp()

    @mock.patch('zipfile.is_zipfile')
    @mock.patch('os.path.getsize')
    def test__get_package_file_package(self, getsize_mock, is_zipfile_mock):
        is_zipfile_mock.return_value = True
        getsize_mock.return_value = function.MAX_ZIP_SIZE

        package_path = '/PATH/TO/FAKE_PACKAGE.ZIP'

        ret = function._get_package_file(package_path=package_path)

        self.assertEqual(package_path, ret)

        is_zipfile_mock.assert_called_once_with(package_path)
        getsize_mock.assert_called_once_with(package_path)

    @mock.patch('zipfile.is_zipfile')
    def test__get_package_file_package_not_a_valid_zip_file(self,
                                                            is_zipfile_mock):
        is_zipfile_mock.return_value = False

        package_path = '/PATH/TO/NOT_A_ZIP_FILE.py'

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r"^Package %s is not a valid ZIP file\.$" % package_path,
            function._get_package_file, package_path=package_path)

        is_zipfile_mock.assert_called_once_with(package_path)

    @mock.patch('zipfile.is_zipfile')
    @mock.patch('os.path.getsize')
    def test__get_package_file_package_size_exceed(self, getsize_mock,
                                                   is_zipfile_mock):
        is_zipfile_mock.return_value = True
        getsize_mock.return_value = function.MAX_ZIP_SIZE + 1

        package_path = '/PATH/TO/FAKE_PACKAGE.ZIP'

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Package file size must be no more than %sM\.$' % (
                function.MAX_ZIP_SIZE / 1024 / 1024),
            function._get_package_file, package_path=package_path)

        is_zipfile_mock.assert_called_once_with(package_path)
        getsize_mock.assert_called_once_with(package_path)

    @mock.patch('os.path.isfile')
    @mock.patch('tempfile.gettempdir')
    @mock.patch('zipfile.ZipFile')
    @mock.patch('os.path.getsize')
    def test__get_package_file_file(self, getsize_mock, ZipFileClassMock,
                                    gettempdir_mock, isfile_mock):
        isfile_mock.return_value = True
        fake_tempdir_path = '/FAKE/TEMP_DIR'
        gettempdir_mock.return_value = fake_tempdir_path
        fake_zf = mock.Mock()
        ZipFileClassMock.return_value = fake_zf
        getsize_mock.return_value = function.MAX_ZIP_SIZE

        base_name = 'FAKE_FILE'
        extension = '.py'
        file_path = '/PATH/TO/%s%s' % (base_name, extension)
        zip_file_path = '%s/%s.zip' % (fake_tempdir_path, base_name)

        ret = function._get_package_file(file_path=file_path)

        self.assertEqual(zip_file_path, ret)

        isfile_mock.assert_called_once_with(file_path)
        ZipFileClassMock.assert_called_once_with(zip_file_path, mode='w')
        fake_zf.write.assert_called_once_with(
            file_path, '%s%s' % (base_name, extension),
            compress_type=zipfile.ZIP_STORED)
        fake_zf.close.assert_called_once_with()
        getsize_mock.assert_called_once_with(zip_file_path)

    @mock.patch('os.path.isfile')
    def test__get_package_file_file_not_exist(self, isfile_mock):
        isfile_mock.return_value = False

        file_path = '/PATH/TO/A_NON_EXIST_FILE.py'

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r"^File %s not exist\.$" % file_path,
            function._get_package_file, file_path=file_path)

        isfile_mock.assert_called_once_with(file_path)

    @mock.patch('os.path.isfile')
    @mock.patch('tempfile.gettempdir')
    @mock.patch('zipfile.ZipFile')
    @mock.patch('os.path.getsize')
    def test__get_package_file_file_zipped_size_exceed(self, getsize_mock,
                                                       ZipFileClassMock,
                                                       gettempdir_mock,
                                                       isfile_mock):
        isfile_mock.return_value = True
        fake_tempdir_path = '/FAKE/TEMP_DIR'
        gettempdir_mock.return_value = fake_tempdir_path
        fake_zf = mock.Mock()
        ZipFileClassMock.return_value = fake_zf
        getsize_mock.return_value = function.MAX_ZIP_SIZE + 1

        base_name = 'FAKE_FILE'
        extension = '.py'
        file_path = '/PATH/TO/%s%s' % (base_name, extension)
        zip_file_path = '%s/%s.zip' % (fake_tempdir_path, base_name)

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Package file size must be no more than %sM\.$' % (
                function.MAX_ZIP_SIZE / 1024 / 1024),
            function._get_package_file, file_path=file_path)

        isfile_mock.assert_called_once_with(file_path)
        ZipFileClassMock.assert_called_once_with(zip_file_path, mode='w')
        fake_zf.write.assert_called_once_with(
            file_path, '%s%s' % (base_name, extension),
            compress_type=zipfile.ZIP_STORED)
        fake_zf.close.assert_called_once_with()
        getsize_mock.assert_called_once_with(zip_file_path)


class TestFunction(fakes.TestQinlingClient):

    def setUp(self):
        super(TestFunction, self).setUp()
        # Get a shortcut
        self.client = self.app.client_manager.function_engine

        self.columns = base.FUNCTION_COLUMNS
        self.data = []

        self._runtimes = fakes.FakeRuntime.create_runtimes(count=2)
        self._functions = fakes.FakeFunction.create_functions(count=3)
        for f in self._functions:
            self.data.append((f.id, f.name, f.description, f.count, f.code,
                              f.runtime_id, f.entry, f.project_id,
                              f.created_at, f.updated_at,
                              f.cpu, f.memory_size, f.timeout))


class TestListFunction(TestFunction):

    def setUp(self):
        super(TestListFunction, self).setUp()
        self.cmd = function.List(self.app, None)

        self.columns = [c.capitalize() for c in base.FUNCTION_COLUMNS]

        self.client.functions.list = mock.Mock(return_value=self._functions)

    def test_function_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.client.functions.list.assert_called_once_with()
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_function_list_with_filter(self):
        arglist = ['--filter', 'name=has:function',
                   '--filter', 'cpu=lt:200']
        verifylist = [
            ('filters', ['name=has:function', 'cpu=lt:200'])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.client.functions.list.assert_called_once_with(
            name='has:function', cpu='lt:200'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_function_list_with_invalid_filter(self):
        arglist = ['--filter', 'name']
        verifylist = [
            ('filters', ['name'])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            ValueError,
            '^Invalid filter: name$',
            self.cmd.take_action, parsed_args
        )


class TestCreateFunction(TestFunction):

    def setUp(self):
        super(TestCreateFunction, self).setUp()
        self.cmd = function.Create(self.app, None)
        self.runtime_id = self._runtimes[0].id
        self.runtime_name = self._runtimes[0].name
        self.function_name = 'FAKE_FUNCTION_NAME'
        self.function_entry = 'fake_module.fake_method'
        self.file_path = '/PATH/TO/functon.py'
        self.package_path = '/FAKE_TEMP_DIR/ZIPPED_function.zip'
        self.container_name = 'FAKE_SWIFT_CONTAINER'
        self.object_name = 'FAKE_SWIFT_OBJECT'
        self.image = 'FAKE_IMAGE'
        self.cpu = 200
        self.memory_size = 64 * 1024 * 1024
        self.timeout = 10
        # The arguments below are used in every create call despite of the
        # function type.
        self.base_call_arguments_default_values = {
            'name': None,
            'entry': None,
            'cpu': None,
            'memory_size': None,
            'timeout': 5
        }
        self.base_call_arguments_passed_values = {
            'name': self.function_name,
            'entry': self.function_entry,
            'cpu': self.cpu,
            'memory_size': self.memory_size,
            'timeout': self.timeout
        }

    def _create_fake_function(self, attrs=None):
        # Allow to fake different create results
        f = fakes.FakeFunction.create_one_function(attrs)
        self.client.functions.create = mock.Mock(return_value=f)
        data = (f.id, f.name, f.description, f.count, f.code,
                f.runtime_id, f.entry, f.project_id,
                f.created_at, f.updated_at,
                f.cpu, f.memory_size, f.timeout)
        return data

    def _get_verify_list(self, runtime=None, name=None, entry=None,
                         file_path=None, package_path=None,
                         container_name=None, object_name=None,
                         image=None, cpu=None, memory_size=None, timeout=5):
        return [
            ('runtime', runtime),
            ('name', name),
            ('entry', entry),
            ('file', file_path),
            ('package', package_path),
            ('container', container_name),
            ('object', object_name),
            ('image', image),
            ('cpu', cpu),
            ('memory_size', memory_size),
            ('timeout', timeout)
        ]

    def _get_verify_list_with_passed_value(self, **kwargs):
        kwargs.update({'name': self.function_name,
                       'entry': self.function_entry,
                       'cpu': self.cpu,
                       'memory_size': self.memory_size,
                       'timeout': self.timeout})
        return self._get_verify_list(**kwargs)

    def test_function_create_no_option(self):
        arglist = []
        verifylist = self._get_verify_list()

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            '^Cannot create function with the parameters given.+',
            self.cmd.take_action, parsed_args)

    def test_function_create_file_and_package_mutually_exclusive(self):
        # --file and --package are mutually exclusive
        arglist = ['--file', self.file_path, '--package', self.package_path]
        verifylist = self._get_verify_list(file_path=self.file_path,
                                           package_path=self.package_path)

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    @mock.patch('qinlingclient.osc.v1.function._get_package_file')
    @mock.patch('qinlingclient.utils.md5')
    @mock.patch('qinlingclient.osc.v1.function.open')
    @mock.patch('os.remove')
    def test_function_create_package_required_options(
            self, remove_mock, open_mock, md5_mock, _get_package_file_mock):
        """Create a package type function.

        1. file param is specified,
        2. use runtime id,
        3. all other params except the required ones are not set.
        """
        md5sum = 'FAKE_MD5_SUM'
        code = {'source': 'package', 'md5sum': md5sum}
        attrs = {'runtime_id': self.runtime_id, 'code': code}
        created_data = self._create_fake_function(attrs)

        _get_package_file_mock.return_value = self.package_path
        md5_mock.return_value = md5sum
        package_file = mock.Mock()
        open_mock.return_value.__enter__.return_value = package_file

        arglist = ['--runtime', self.runtime_id, '--file', self.file_path]
        verifylist = self._get_verify_list(runtime=self.runtime_id,
                                           file_path=self.file_path)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        call_arguments = {'runtime': self.runtime_id,
                          'code': code,
                          'package': package_file}
        call_arguments.update(self.base_call_arguments_default_values)
        self.client.functions.create.assert_called_once_with(**call_arguments)
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

        _get_package_file_mock.assert_called_once_with(None, self.file_path)
        md5_mock.assert_called_once_with(file=self.package_path)
        open_mock.assert_called_once_with(self.package_path, 'rb')
        remove_mock.assert_called_once_with(self.package_path)

    @mock.patch('qinlingclient.osc.v1.function._get_package_file')
    @mock.patch('qinlingclient.utils.md5')
    @mock.patch('qinlingclient.osc.v1.function.open')
    @mock.patch('os.remove')
    def test_function_create_package_all_options(
            self, remove_mock, open_mock, md5_mock, _get_package_file_mock):
        """Create a package type function.

        1. package param is specified
        2. use runtime name,
        3. if acceptable, all options are specified with some value, so this
           also tests that package type is taking precedence of other types.
        """
        md5sum = 'FAKE_MD5_SUM'
        code = {'source': 'package', 'md5sum': md5sum}
        attrs = {
            'name': self.function_name,
            'code': code,
            'runtime_id': self.runtime_id,
            'entry': self.function_entry,
            'cpu': self.cpu,
            'memory_size': self.memory_size,
            'timeout': self.timeout
        }
        created_data = self._create_fake_function(attrs)

        _get_package_file_mock.return_value = self.package_path
        md5_mock.return_value = md5sum
        package_file = mock.Mock()
        open_mock.return_value.__enter__.return_value = package_file
        # Use to find the runtime id with its name
        self.client.runtimes.find.return_value = self._runtimes[0]

        arglist = [
            '--runtime', self.runtime_name,
            '--name', self.function_name,
            '--entry', self.function_entry,
            '--package', self.package_path,
            '--container', self.container_name,
            '--object', self.object_name,
            '--image', self.image,
            '--cpu', str(self.cpu),
            '--memory-size', str(self.memory_size),
            '--timeout', str(self.timeout)
        ]
        verifylist = self._get_verify_list_with_passed_value(
            runtime=self.runtime_name,
            package_path=self.package_path,
            container_name=self.container_name,
            object_name=self.object_name,
            image=self.image,
            cpu=self.cpu,
            memory_size=self.memory_size,
        )

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        call_arguments = {'runtime': self.runtime_id,
                          'code': code,
                          'package': package_file}
        call_arguments.update(self.base_call_arguments_passed_values)
        self.client.functions.create.assert_called_once_with(**call_arguments)
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

        self.client.runtimes.find.assert_called_once_with(
            name=self.runtime_name)
        _get_package_file_mock.assert_called_once_with(self.package_path, None)
        md5_mock.assert_called_once_with(file=self.package_path)
        open_mock.assert_called_once_with(self.package_path, 'rb')

    def test_function_create_package_runtime_needed(self):
        """Create a package type function without runtime specified."""

        arglist = ['--file', self.file_path]
        verifylist = self._get_verify_list(file_path=self.file_path)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Runtime needs to be specified for package type function\.$',
            self.cmd.take_action, parsed_args)

    def test_function_create_swift_required_options(self):
        """Create a swift type function.

        1. use runtime id,
        2. all other params except the required ones are not set.
        """
        code = {
            'source': 'swift',
            'swift': {
                'container': self.container_name,
                'object': self.object_name
            }
        }
        attrs = {'runtime_id': self.runtime_id, 'code': code}
        created_data = self._create_fake_function(attrs)

        arglist = ['--runtime', self.runtime_id,
                   '--container', self.container_name,
                   '--object', self.object_name]
        verifylist = self._get_verify_list(runtime=self.runtime_id,
                                           container_name=self.container_name,
                                           object_name=self.object_name)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        call_arguments = {'runtime': self.runtime_id, 'code': code}
        call_arguments.update(self.base_call_arguments_default_values)
        self.client.functions.create.assert_called_once_with(**call_arguments)
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

    def test_function_create_swift_all_options(self):
        """Create a swift type function.

        1. use runtime name,
        2. if acceptable, all options are specified with some value, so this
           also tests that swift type is taking precedence of image type.
        """

        code = {
            'source': 'swift',
            'swift': {
                'container': self.container_name,
                'object': self.object_name
            }
        }
        attrs = {
            'name': self.function_name,
            'code': code,
            'runtime_id': self.runtime_id,
            'entry': self.function_entry,
            'cpu': self.cpu,
            'memory_size': self.memory_size,
            'timeout': self.timeout
        }
        created_data = self._create_fake_function(attrs)

        # Use to find the runtime id with its name
        self.client.runtimes.find.return_value = self._runtimes[0]

        arglist = [
            '--runtime', self.runtime_name,
            '--name', self.function_name,
            '--entry', self.function_entry,
            '--container', self.container_name,
            '--object', self.object_name,
            '--image', self.image,
            '--cpu', str(self.cpu),
            '--memory-size', str(self.memory_size),
            '--timeout', str(self.timeout)
        ]
        verifylist = self._get_verify_list_with_passed_value(
            runtime=self.runtime_name,
            container_name=self.container_name,
            object_name=self.object_name,
            image=self.image
        )

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        call_arguments = {'runtime': self.runtime_id, 'code': code}
        call_arguments.update(self.base_call_arguments_passed_values)
        self.client.functions.create.assert_called_once_with(**call_arguments)
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

        self.client.runtimes.find.assert_called_once_with(
            name=self.runtime_name)

    def test_function_create_swift_container_needed(self):
        """Create a swift type function with only object given."""

        arglist = ['--object', self.object_name]
        verifylist = self._get_verify_list(object_name=self.object_name)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Container name and object name need to be specified\.$',
            self.cmd.take_action, parsed_args)

    def test_function_create_swift_object_needed(self):
        """Create a swift type function with only container given."""

        arglist = ['--container', self.container_name]
        verifylist = self._get_verify_list(container_name=self.container_name)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Container name and object name need to be specified\.$',
            self.cmd.take_action, parsed_args)

    def test_function_create_swift_runtime_needed(self):
        """Create a swift type function without runtime given."""

        arglist = ['--container', self.container_name,
                   '--object', self.object_name]
        verifylist = self._get_verify_list(container_name=self.container_name,
                                           object_name=self.object_name)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Runtime needs to be specified for swift type function\.$',
            self.cmd.take_action, parsed_args)

    def test_function_create_image_required_options(self):
        """Create a image type function with only required options."""

        code = {
            'source': 'image',
            'image': self.image
        }
        attrs = {'code': code}
        created_data = self._create_fake_function(attrs)

        arglist = ['--image', self.image]
        verifylist = self._get_verify_list(image=self.image)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        call_arguments = {'code': code}
        call_arguments.update(self.base_call_arguments_default_values)
        self.client.functions.create.assert_called_once_with(**call_arguments)
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

    def test_function_create_image_all_options(self):
        code = {
            'source': 'image',
            'image': self.image
        }
        attrs = {
            'name': self.function_name,
            'code': code,
            'entry': self.function_entry,
            'cpu': self.cpu,
            'memory_size': self.memory_size,
            'timeout': self.timeout
        }
        created_data = self._create_fake_function(attrs)

        arglist = [
            '--name', self.function_name,
            '--entry', self.function_entry,
            '--image', self.image,
            '--cpu', str(self.cpu),
            '--memory-size', str(self.memory_size),
            '--timeout', str(self.timeout)
        ]
        verifylist = self._get_verify_list_with_passed_value(image=self.image)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        call_arguments = {'code': code}
        call_arguments.update(self.base_call_arguments_passed_values)
        self.client.functions.create.assert_called_once_with(**call_arguments)
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

    def test_function_create_timeout_invalid(self):
        arglist = ['--timeout', '-1']
        verifylist = self._get_verify_list(timeout=-1)

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestDeleteFunction(TestFunction):

    def setUp(self):
        super(TestDeleteFunction, self).setUp()
        self.cmd = function.Delete(self.app, None)
        self.client.functions.delete = mock.Mock(return_value=None)

    def test_function_delete_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_delete_one(self):
        function_id = self._functions[0].id
        arglist = [function_id]
        verifylist = [('function', [function_id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.functions.delete.assert_called_once_with(function_id)

    @mock.patch('qinlingclient.utils.find_resource_id_by_name')
    def test_function_delete_one_by_name(self, mock_find):
        name = self._functions[0].name
        id = self._functions[0].id
        mock_find.return_value = id
        arglist = [name]
        verifylist = [('function', [name])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.functions.delete.assert_called_once_with(id)

    def test_function_delete_multiple(self):
        function_ids = [r.id for r in self._functions]
        arglist = function_ids
        verifylist = [('function', function_ids)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        calls = [mock.call(f_id) for f_id in function_ids]
        self.assertEqual(len(function_ids),
                         self.client.functions.delete.call_count)
        self.client.functions.delete.assert_has_calls(calls)

    @mock.patch('qinlingclient.utils.find_resource_id_by_name')
    def test_function_delete_multiple_with_names(self, mock_find):
        function_ids = [r.id for r in self._functions[:-1]]
        function_names = [self._functions[-1].name]
        mock_find.return_value = self._functions[-1].id
        arglist = function_ids + function_names
        verifylist = [('function', function_ids + function_names)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)

        expected = [mock.call(f_id) for f_id in function_ids]
        expected.append(mock.call(self._functions[-1].id))

        self.client.functions.delete.assert_has_calls(expected)

    def test_function_delete_multiple_exception(self):
        function_ids = [r.id for r in self._functions]
        arglist = function_ids
        verifylist = [('function', function_ids)]

        self.client.functions.delete = mock.Mock(side_effect=[
            None, RuntimeError, None
        ])

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Unable to delete the specified function\(s\)\.$',
            self.cmd.take_action, parsed_args)

        # The second deletion failed, but the third is done normally
        calls = [mock.call(f_id) for f_id in function_ids]
        self.assertEqual(len(function_ids),
                         self.client.functions.delete.call_count)
        self.client.functions.delete.assert_has_calls(calls)


class TestShowFunction(TestFunction):

    def setUp(self):
        super(TestShowFunction, self).setUp()
        self.cmd = function.Show(self.app, None)
        self.client.functions.get = mock.Mock(return_value=self._functions[0])

    def test_function_show_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_show(self):
        function_id = self._functions[0].id
        arglist = [function_id]
        verifylist = [('function', function_id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.functions.get.assert_called_once_with(function_id)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data[0], data)

    @mock.patch("qinlingclient.utils.find_resource_id_by_name")
    def test_function_show_by_name(self, mock_find):
        name = self._functions[0].name
        id = self._functions[0].id
        mock_find.return_value = id
        arglist = [name]
        verifylist = [('function', name)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        mock_find.assert_called_once_with(self.client.functions, name)
        self.client.functions.get.assert_called_once_with(id)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data[0], data)


class TestUpdateFunction(TestFunction):

    def setUp(self):
        super(TestUpdateFunction, self).setUp()
        self.cmd = function.Update(self.app, None)
        self.function_id = self._functions[0].id
        self.function_name = 'FAKE_FUNCTION_NAME'
        self.function_description = 'This is a updated function.'
        self.function_entry = 'fake_module.fake_method'
        self.file_path = '/PATH/TO/functon.py'
        self.package_path = '/FAKE_TEMP_DIR/ZIPPED_function.zip'
        self.container_name = 'FAKE_SWIFT_CONTAINER'
        self.object_name = 'FAKE_SWIFT_OBJECT'
        self.image = 'FAKE_IMAGE'
        self.cpu = 200
        self.memory_size = 64 * 1024 * 1024
        self.timeout = 10
        # The arguments below are used in every update despite of the
        # function type.
        self.base_call_arguments_default_values = {
            'name': None,
            'description': None,
            'entry': None,
            'cpu': None,
            'memory_size': None,
            'timeout': None,
        }
        self.base_call_arguments_passed_values = {
            'name': self.function_name,
            'description': self.function_description,
            'entry': self.function_entry,
            'cpu': self.cpu,
            'memory_size': self.memory_size,
            'timeout': self.timeout
        }

    def _update_fake_function(self, attrs=None):
        # Allow to fake different update results
        f = fakes.FakeFunction.create_one_function(attrs)
        self.client.functions.update = mock.Mock(return_value=f)
        data = (f.id, f.name, f.description, f.count, f.code,
                f.runtime_id, f.entry, f.project_id,
                f.created_at, f.updated_at,
                f.cpu, f.memory_size, f.timeout)
        return data

    def _get_verify_list(self, function_id=None, name=None, description=None,
                         entry=None, file_path=None, package_path=None,
                         container_name=None, object_name=None,
                         cpu=None, memory_size=None, timeout=None):
        return [
            ('function', function_id),
            ('name', name),
            ('description', description),
            ('entry', entry),
            ('file', file_path),
            ('package', package_path),
            ('container', container_name),
            ('object', object_name),
            ('cpu', cpu),
            ('memory_size', memory_size),
            ('timeout', timeout)
        ]

    def _get_verify_list_with_passed_value(self, **kwargs):
        kwargs.update({'name': self.function_name,
                       'description': self.function_description,
                       'entry': self.function_entry,
                       'cpu': self.cpu,
                       'memory_size': self.memory_size,
                       'timeout': self.timeout})
        return self._get_verify_list(**kwargs)

    def test_function_update_no_option(self):
        arglist = []
        verifylist = self._get_verify_list()

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_update_required_options(self):
        """Update a function.

        Do nothing as only the function_id is specified.
        """
        attrs = {'id': self.function_id}
        updated_data = self._update_fake_function(attrs)

        arglist = [self.function_id]
        verifylist = self._get_verify_list(function_id=self.function_id)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        call_arguments = {'code': None}
        call_arguments.update(self.base_call_arguments_default_values)
        self.client.functions.update.assert_called_once_with(
            self.function_id, **call_arguments)
        self.assertEqual(self.columns, columns)
        self.assertEqual(updated_data, data)

    @mock.patch('qinlingclient.utils.find_resource_id_by_name')
    def test_function_update_by_name(self, mock_find):
        attrs = {'name': self.function_name, 'timeout': self.timeout}
        updated_data = self._update_fake_function(attrs)
        mock_find.return_value = self.function_id

        arglist = [self.function_name, '--timeout', str(self.timeout)]
        verifylist = self._get_verify_list(function_id=self.function_name,
                                           timeout=self.timeout)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        mock_find.assert_called_once_with(self.client.functions,
                                          self.function_name)

        call_arguments = {'code': None}
        call_arguments.update(self.base_call_arguments_default_values)
        call_arguments.update({'timeout': self.timeout})
        self.client.functions.update.assert_called_once_with(
            self.function_id, **call_arguments)

        self.assertEqual(self.columns, columns)
        self.assertEqual(updated_data, data)

    def test_function_update_code_not_updated(self):
        # All other options except code are updated.
        attrs = {'id': self.function_id,
                 'name': self.function_name,
                 'description': self.function_description,
                 'entry': self.function_entry,
                 'cpu': self.cpu,
                 'memory_size': self.memory_size,
                 'timeout': self.timeout}
        updated_data = self._update_fake_function(attrs)

        arglist = [
            self.function_id,
            '--name', self.function_name,
            '--description', self.function_description,
            '--entry', self.function_entry,
            '--cpu', str(self.cpu),
            '--memory-size', str(self.memory_size),
            '--timeout', str(self.timeout)
        ]
        verifylist = self._get_verify_list_with_passed_value(
            function_id=self.function_id)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        call_arguments = {'code': None}
        call_arguments.update(self.base_call_arguments_passed_values)
        self.client.functions.update.assert_called_once_with(
            self.function_id, **call_arguments)
        self.assertEqual(self.columns, columns)
        self.assertEqual(updated_data, data)

    def test_function_update_file_and_package_mutually_exclusive(self):
        # --file and --package are mutually exclusive
        arglist = [self.function_id,
                   '--file', self.file_path, '--package', self.package_path]
        verifylist = self._get_verify_list(function_id=self.function_id,
                                           file_path=self.file_path,
                                           package_path=self.package_path)

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    @mock.patch('qinlingclient.osc.v1.function._get_package_file')
    @mock.patch('qinlingclient.osc.v1.function.open')
    def test_function_update_package_required_options(
            self, open_mock, _get_package_file_mock):
        """Update a package type function.

        1. use --file to update the code,
        2. only required options are specified.
        """
        code = {'source': 'package'}
        attrs = {'id': self.function_id, 'code': code}
        updated_data = self._update_fake_function(attrs)

        _get_package_file_mock.return_value = self.package_path
        package_file = mock.Mock()
        open_mock.return_value.__enter__.return_value = package_file

        arglist = [self.function_id, '--file', self.file_path]
        verifylist = self._get_verify_list(function_id=self.function_id,
                                           file_path=self.file_path)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        call_arguments = {'code': code, 'package': package_file}
        call_arguments.update(self.base_call_arguments_default_values)
        self.client.functions.update.assert_called_once_with(
            self.function_id, **call_arguments)
        self.assertEqual(self.columns, columns)
        self.assertEqual(updated_data, data)

        _get_package_file_mock.assert_called_once_with(None, self.file_path)
        open_mock.assert_called_once_with(self.package_path, 'rb')

    @mock.patch('qinlingclient.osc.v1.function._get_package_file')
    @mock.patch('qinlingclient.osc.v1.function.open')
    def test_function_update_package_all_options(
            self, open_mock, _get_package_file_mock):
        """Update a package type function.

        1. use --package to update the code,
        2. all the options are specified if acceptable, this also tests
           that package type is taking precedence of the swift type.
        """
        code = {'source': 'package'}
        attrs = {'id': self.function_id, 'code': code}
        updated_data = self._update_fake_function(attrs)

        _get_package_file_mock.return_value = self.package_path
        package_file = mock.Mock()
        open_mock.return_value.__enter__.return_value = package_file

        arglist = [self.function_id, '--package', self.package_path,
                   '--name', self.function_name,
                   '--description', self.function_description,
                   '--entry', self.function_entry,
                   '--container', self.container_name,
                   '--object', self.object_name,
                   '--cpu', str(self.cpu),
                   '--memory-size', str(self.memory_size),
                   '--timeout', str(self.timeout)]
        verifylist = self._get_verify_list_with_passed_value(
            function_id=self.function_id,
            package_path=self.package_path,
            container_name=self.container_name,
            object_name=self.object_name)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        call_arguments = {'code': code, 'package': package_file}
        call_arguments.update(self.base_call_arguments_passed_values)
        self.client.functions.update.assert_called_once_with(
            self.function_id, **call_arguments)
        self.assertEqual(self.columns, columns)
        self.assertEqual(updated_data, data)

        _get_package_file_mock.assert_called_once_with(self.package_path, None)
        open_mock.assert_called_once_with(self.package_path, 'rb')

    def test_function_update_swift_required_options(self):
        """Update a swift type function.

        1. only use --container to update the code.
        """
        code = {
            'source': 'swift',
            'swift': {
                'container': self.container_name,
                'object': "origin_obj"
            }
        }
        attrs = {'id': self.function_id, 'code': code}
        updated_data = self._update_fake_function(attrs)

        arglist = [self.function_id, '--container', self.container_name]
        verifylist = self._get_verify_list(function_id=self.function_id,
                                           container_name=self.container_name)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        code = {
            'source': 'swift',
            'swift': {
                'container': self.container_name,
            }
        }
        call_arguments = {'code': code}
        call_arguments.update(self.base_call_arguments_default_values)
        self.client.functions.update.assert_called_once_with(
            self.function_id, **call_arguments)
        self.assertEqual(self.columns, columns)
        self.assertEqual(updated_data, data)

    def test_function_update_swift_all_options(self):
        """Update a swift type function.

        1. only both --container and --object to update the code,
        2. all other options are specified if appropriate.
        """
        code = {
            'source': 'swift',
            'swift': {
                'container': self.container_name,
                'object': self.object_name
            }
        }
        attrs = {'id': self.function_id, 'code': code}
        updated_data = self._update_fake_function(attrs)

        arglist = [self.function_id,
                   '--container', self.container_name,
                   '--object', self.object_name,
                   '--name', self.function_name,
                   '--description', self.function_description,
                   '--entry', self.function_entry,
                   '--cpu', str(self.cpu),
                   '--memory-size', str(self.memory_size),
                   '--timeout', str(self.timeout)]
        verifylist = self._get_verify_list_with_passed_value(
            function_id=self.function_id,
            container_name=self.container_name,
            object_name=self.object_name)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        call_arguments = {'code': code}
        call_arguments.update(self.base_call_arguments_passed_values)
        self.client.functions.update.assert_called_once_with(
            self.function_id, **call_arguments)
        self.assertEqual(self.columns, columns)
        self.assertEqual(updated_data, data)

    def test_function_update_timeout_invalid(self):
        arglist = [self.function_id, '--timeout', '-1']
        verifylist = self._get_verify_list(function_id=self.function_id,
                                           timeout=-1)

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestDetachFunction(TestFunction):

    def setUp(self):
        super(TestDetachFunction, self).setUp()
        self.cmd = function.Detach(self.app, None)
        self.client.functions.detach = mock.Mock(return_value=None)

    def test_function_detach_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_detach(self):
        function_id = self._functions[0].id
        arglist = [function_id]
        verifylist = [('function', function_id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.functions.detach.assert_called_once_with(function_id)

    def test_function_detach_exception(self):
        function_id = self._functions[0].id
        self.client.functions.detach = mock.Mock(side_effect=RuntimeError)

        arglist = [function_id]
        verifylist = [('function', function_id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Unable to detach the specified function\.$',
            self.cmd.take_action, parsed_args)

        self.client.functions.detach.assert_called_once_with(function_id)


class TestDownloadFunction(TestFunction):

    def setUp(self):
        super(TestDownloadFunction, self).setUp()
        self.cmd = function.Download(self.app, None)
        self.raw_data = 'RAW_DATA'
        self.cwd = '/FAKE/CWD'
        res = mock.Mock(raw=self.raw_data)
        self.client.functions.get = mock.Mock(return_value=res)

    def test_function_download_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    @mock.patch('os.getcwd')
    @mock.patch('qinlingclient.osc.v1.function.open')
    @mock.patch('shutil.copyfileobj')
    def test_function_download_required_options(self, copyfile_mock, open_mock,
                                                getcwd_mock):
        function_id = self._functions[0].id

        getcwd_mock.return_value = self.cwd
        target = mock.Mock()
        open_mock.return_value.__enter__.return_value = target

        arglist = [function_id]
        verifylist = [('function', function_id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.functions.get.assert_called_once_with(function_id,
                                                          download=True)
        open_mock.assert_called_once_with(
            '%s/%s.zip' % (self.cwd, function_id), 'wb')
        copyfile_mock.assert_called_once_with(self.raw_data, target)

    @mock.patch('os.getcwd')
    @mock.patch('qinlingclient.osc.v1.function.open')
    @mock.patch('shutil.copyfileobj')
    @mock.patch('qinlingclient.utils.find_resource_id_by_name')
    def test_function_download_by_name(self, find_mock, copyfile_mock,
                                       open_mock, getcwd_mock):
        name = self._functions[0].name
        id = self._functions[0].id
        find_mock.return_value = id
        getcwd_mock.return_value = self.cwd
        target = mock.Mock()
        open_mock.return_value.__enter__.return_value = target

        arglist = [name]
        verifylist = [('function', name)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        find_mock.assert_called_once_with(self.client.functions, name)
        self.client.functions.get.assert_called_once_with(id,
                                                          download=True)
        open_mock.assert_called_once_with(
            '%s/%s.zip' % (self.cwd, id), 'wb')
        copyfile_mock.assert_called_once_with(self.raw_data, target)

    @mock.patch('os.getcwd')
    @mock.patch('qinlingclient.osc.v1.function.open')
    @mock.patch('shutil.copyfileobj')
    def test_function_download_all_options(self, copyfile_mock, open_mock,
                                           getcwd_mock):
        function_id = self._functions[0].id
        output_name = 'output.zip'

        getcwd_mock.return_value = self.cwd
        target = mock.Mock()
        open_mock.return_value.__enter__.return_value = target

        arglist = [function_id, '--output', output_name]
        verifylist = [('function', function_id), ('output', output_name)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.functions.get.assert_called_once_with(function_id,
                                                          download=True)
        open_mock.assert_called_once_with(
            '%s/%s' % (self.cwd, output_name), 'wb')
        copyfile_mock.assert_called_once_with(self.raw_data, target)

    @mock.patch('os.getcwd')
    @mock.patch('qinlingclient.osc.v1.function.open')
    @mock.patch('shutil.copyfileobj')
    def test_function_download_abs_path(self, copyfile_mock, open_mock,
                                        getcwd_mock):
        function_id = self._functions[0].id
        output_name = '/ABS_PATH/output.zip'

        getcwd_mock.return_value = self.cwd
        target = mock.Mock()
        open_mock.return_value.__enter__.return_value = target

        arglist = [function_id, '-o', output_name]
        verifylist = [('function', function_id), ('output', output_name)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.functions.get.assert_called_once_with(function_id,
                                                          download=True)
        open_mock.assert_called_once_with(output_name, 'wb')
        copyfile_mock.assert_called_once_with(self.raw_data, target)


class TestScaleupFunction(TestFunction):

    def setUp(self):
        super(TestScaleupFunction, self).setUp()
        self.cmd = function.Scaleup(self.app, None)
        self.client.functions.scaleup = mock.Mock(return_value=None)

    def test_function_scaleup_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_scaleup_required_options(self):
        function_id = self._functions[0].id
        arglist = [function_id]
        verifylist = [('function', function_id), ('count', 1)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.functions.scaleup.assert_called_once_with(function_id, 1)

    def test_function_scaleup_all_options(self):
        function_id = self._functions[0].id
        count = 3
        arglist = [function_id, '--count', str(count)]
        verifylist = [('function', function_id), ('count', count)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.functions.scaleup.assert_called_once_with(function_id,
                                                              count)

    def test_function_scaleup_worker_count_not_int(self):
        function_id = self._functions[0].id
        arglist = [function_id, '--count', 'NOT_INTEGER']
        verifylist = [('function_id', function_id), ('count', 'NOT_INTEGER')]

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Worker count must be a positive integer\.$',
            self.check_parser,
            self.cmd, arglist, verifylist)

    def test_function_scaleup_worker_count_zero(self):
        function_id = self._functions[0].id
        arglist = [function_id, '--count', '0']
        verifylist = [('function_id', function_id), ('count', 0)]

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Worker count must be a positive integer\.$',
            self.check_parser,
            self.cmd, arglist, verifylist)

    def test_function_scaleup_exception(self):
        function_id = self._functions[0].id
        self.client.functions.scaleup = mock.Mock(side_effect=RuntimeError)

        arglist = [function_id]
        verifylist = [('function', function_id), ('count', 1)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Unable to scale up the specified function\.$',
            self.cmd.take_action, parsed_args)

        self.client.functions.scaleup.assert_called_once_with(function_id, 1)


class TestScaledownFunction(TestFunction):

    def setUp(self):
        super(TestScaledownFunction, self).setUp()
        self.cmd = function.Scaledown(self.app, None)
        self.client.functions.scaledown = mock.Mock(return_value=None)

    def test_function_scaledown_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_scaledown_required_options(self):
        function_id = self._functions[0].id
        arglist = [function_id]
        verifylist = [('function', function_id), ('count', 1)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.functions.scaledown.assert_called_once_with(function_id, 1)

    def test_function_scaledown_all_options(self):
        function_id = self._functions[0].id
        count = 3
        arglist = [function_id, '--count', str(count)]
        verifylist = [('function', function_id), ('count', 3)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.functions.scaledown.assert_called_once_with(function_id,
                                                                count)

    def test_function_scaledown_exception(self):
        function_id = self._functions[0].id
        self.client.functions.scaledown = mock.Mock(side_effect=RuntimeError)

        arglist = [function_id]
        verifylist = [('function', function_id), ('count', 1)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Unable to scale down the specified function\.$',
            self.cmd.take_action, parsed_args)

        self.client.functions.scaledown.assert_called_once_with(function_id, 1)



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from osc_lib.tests import utils as osc_tests_utils

from qinlingclient.common import exceptions
from qinlingclient.osc.v1 import base
from qinlingclient.osc.v1 import function_version
from qinlingclient.tests.unit.osc.v1 import fakes


class TestFunctionVersion(fakes.TestQinlingClient):

    def setUp(self):
        super(TestFunctionVersion, self).setUp()
        # Get a shortcut
        self.client = self.app.client_manager.function_engine

        self.columns = base.FUNCTION_VERSION_COLUMNS
        self.data = []

        versions = fakes.FakeFunctionVersion.create_function_versions(count=3)
        self._function_versions = versions
        for v in self._function_versions:
            self.data.append((v.id, v.function_id, v.description,
                              v.version_number, v.count, v.project_id,
                              v.created_at, v.updated_at))


class TestListFunctionVersion(TestFunctionVersion):

    def setUp(self):
        super(TestListFunctionVersion, self).setUp()
        self.cmd = function_version.List(self.app, None)

        self.columns = [c.capitalize() for c in base.FUNCTION_VERSION_COLUMNS]

        self.client.function_versions.list = mock.Mock(
            return_value=self._function_versions)

    def test_function_version_list_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_version_list(self):
        function_id = self._function_versions[0].function_id

        arglist = [function_id]
        verifylist = [('function_id', function_id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_versions.list.assert_called_once_with(function_id)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))


class TestCreateFunctionVersion(TestFunctionVersion):

    def setUp(self):
        super(TestCreateFunctionVersion, self).setUp()
        self.cmd = function_version.Create(self.app, None)

    def _create_fake_function_version(self, attrs=None):
        # Allow to fake different create results
        v = fakes.FakeFunctionVersion.create_one_function_version(attrs)
        self.client.function_versions.create = mock.Mock(return_value=v)
        data = (v.id, v.function_id, v.description,
                v.version_number, v.count, v.project_id,
                v.created_at, v.updated_at)
        return data

    def test_function_version_create_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_version_create_required_options(self):
        """Create a function version.

        1. use function_id,
        2. all other params except the required ones are not set.
        """
        function_id = self._function_versions[0].function_id
        attrs = {'function_id': function_id, 'version_number': '2'}
        created_data = self._create_fake_function_version(attrs)

        arglist = [function_id]
        verifylist = [
            ('function', function_id),
            ('description', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_versions.create.assert_called_once_with(
            function_id, description=None
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

    def test_function_version_create_all_options(self):
        """Create a function version.

        1. use function name to find the function_id,
        2. all optional params are specified.
        """
        function = fakes.FakeFunction.create_one_function()
        function_name = function.name
        function_id = function.id
        description = 'This is a new function version.'
        attrs = {'function_id': function_id,
                 'description': description,
                 'version_number': '2'}
        created_data = self._create_fake_function_version(attrs)

        # Use to find the function id with its name
        self.client.functions.find.return_value = function

        arglist = [function_name, '--description', description]
        verifylist = [
            ('function', function_name),
            ('description', description),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_versions.create.assert_called_once_with(
            function_id, description=description
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

        self.client.functions.find.assert_called_once_with(name=function_name)


class TestDeleteFunctionVersion(TestFunctionVersion):

    def setUp(self):
        super(TestDeleteFunctionVersion, self).setUp()
        self.cmd = function_version.Delete(self.app, None)
        self.client.function_versions.delete = mock.Mock(return_value=None)

    def test_function_version_delete_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_version_delete(self):
        function_id = self._function_versions[0].function_id
        version_number = str(self._function_versions[0].version_number)

        arglist = [function_id, version_number]
        verifylist = [
            ('function_id', function_id),
            ('version_number', version_number),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.function_versions.delete.assert_called_once_with(
            function_id, version_number
        )


class TestShowFunctionVersion(TestFunctionVersion):

    def setUp(self):
        super(TestShowFunctionVersion, self).setUp()
        self.cmd = function_version.Show(self.app, None)
        self.client.function_versions.get = mock.Mock(
            return_value=self._function_versions[0]
        )

    def test_function_version_show_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_version_show(self):
        function_id = self._function_versions[0].function_id
        version_number = str(self._function_versions[0].version_number)

        arglist = [function_id, version_number]
        verifylist = [
            ('function_id', function_id),
            ('version_number', version_number),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_versions.get.assert_called_once_with(
            function_id, version_number
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data[0], data)


class TestDetachFunctionVersion(TestFunctionVersion):

    def setUp(self):
        super(TestDetachFunctionVersion, self).setUp()
        self.cmd = function_version.Detach(self.app, None)
        self.client.function_versions.detach = mock.Mock(return_value=None)

    def test_function_version_detach_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_version_detach(self):
        function_id = self._function_versions[0].function_id
        version_number = str(self._function_versions[0].version_number)

        arglist = [function_id, version_number]
        verifylist = [
            ('function_id', function_id),
            ('version_number', version_number),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.function_versions.detach.assert_called_once_with(
            function_id, version_number
        )

    def test_function_version_detach_exception(self):
        function_id = self._function_versions[0].function_id
        version_number = str(self._function_versions[0].version_number)

        self.client.function_versions.detach = mock.Mock(
            side_effect=RuntimeError
        )

        arglist = [function_id, version_number]
        verifylist = [
            ('function_id', function_id),
            ('version_number', version_number),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Unable to detach the specified function version\.$',
            self.cmd.take_action, parsed_args)

        self.client.function_versions.detach.assert_called_once_with(
            function_id, version_number
        )



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from qinlingclient.osc.v1 import base


class List(base.QinlingLister):
    columns = base.WORKER_COLUMNS

    def get_parser(self, prog_name):
        parser = super(List, self).get_parser(prog_name)
        parser.add_argument('function', help='Function ID.')
        return parser

    def _get_resources(self, parsed_args):
        client = self.app.client_manager.function_engine

        return client.function_workers.list(parsed_args.function)



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from osc_lib.tests import utils as osc_tests_utils

from qinlingclient.osc.v1 import base
from qinlingclient.osc.v1 import function_worker
from qinlingclient.tests.unit.osc.v1 import fakes


class TestFunctionWorker(fakes.TestQinlingClient):

    def setUp(self):
        super(TestFunctionWorker, self).setUp()
        # Get a shortcut
        self.client = self.app.client_manager.function_engine

        self.columns = base.WORKER_COLUMNS
        self.data = []

        workers = fakes.FakeFunctionWorker.create_function_workers(count=3)
        self._function_workers = workers
        for w in self._function_workers:
            self.data.append((w.function_id, w.worker_name))


class TestListFunctionWorker(TestFunctionWorker):

    def setUp(self):
        super(TestListFunctionWorker, self).setUp()
        self.cmd = function_worker.List(self.app, None)

        self.columns = [c.capitalize() for c in base.WORKER_COLUMNS]

        self.client.function_workers.list = mock.Mock(
            return_value=self._function_workers)

    def test_function_worker_list_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_function_worker_list(self):
        function_id = self._function_workers[0].function_id

        arglist = [function_id]
        verifylist = [('function', function_id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.client.function_workers.list.assert_called_once_with(function_id)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from osc_lib.command import command
from osc_lib import utils
from oslo_utils import uuidutils

from qinlingclient.osc.v1 import base
from qinlingclient import utils as q_utils


class List(base.QinlingLister):
    columns = base.EXECUTION_COLUMNS
    filtered_columns = base.FILTERED_EXECUTION_COLUMNS

    def _get_resources(self, parsed_args):
        client = self.app.client_manager.function_engine

        return client.function_executions.list(**base.get_filters(parsed_args))


class Create(command.ShowOne):
    columns = base.EXECUTION_COLUMNS

    def get_parser(self, prog_name):
        parser = super(Create, self).get_parser(prog_name)

        parser.add_argument(
            "--function",
            help="Function name or ID.",
        )
        parser.add_argument(
            "--function-version",
            type=int,
            default=0,
            help="Function version number.",
        )
        parser.add_argument(
            "--function-alias",
            help="Function alias which corresponds to a specific function and "
                 "version. When function alias is specified, function and "
                 "function version are not needed.",
        )
        parser.add_argument(
            "--input",
            help="Input for the function.",
        )
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "--sync",
            action='store_true',
            dest='sync',
            default=True,
            help="Run execution synchronously."
        )
        group.add_argument(
            "--async",
            action='store_false',
            dest='sync',
            default=True,
            help="Run execution asynchronously.",
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine

        function_alias = parsed_args.function_alias
        if function_alias:
            function_id = None
            function_version = None
        else:
            function_version = parsed_args.function_version
            function_id = parsed_args.function
            if not uuidutils.is_uuid_like(function_id):
                # Try to find the function id with name
                function_id = q_utils.find_resource_id_by_name(
                    client.functions, function_id)

        execution = client.function_executions.create(
            function_alias=function_alias,
            function_id=function_id,
            function_version=function_version,
            sync=parsed_args.sync,
            input=parsed_args.input
        )

        return self.columns, utils.get_item_properties(execution, self.columns)


class Delete(base.QinlingDeleter):
    def get_parser(self, prog_name):
        parser = super(Delete, self).get_parser(prog_name)

        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "--execution",
            nargs='+',
            help="ID of function execution(s)."
        )
        group.add_argument(
            "--function",
            nargs='+',
            help="ID of function(s)."
        )

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine
        self.delete = client.function_executions.delete
        self.resource = 'execution'

        if parsed_args.execution:
            self.delete_resources(parsed_args.execution)
        elif parsed_args.function:
            for f in parsed_args.function:
                execs = client.function_executions.list(function_id=f)
                ids = [e.id for e in execs]
                self.delete_resources(ids)


class Show(command.ShowOne):
    columns = base.EXECUTION_COLUMNS

    def get_parser(self, prog_name):
        parser = super(Show, self).get_parser(prog_name)
        parser.add_argument('execution', help='Execution ID.')

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine
        execution = client.function_executions.get(parsed_args.execution)

        return self.columns, utils.get_item_properties(execution, self.columns)


class LogShow(command.Command):
    def get_parser(self, prog_name):
        parser = super(LogShow, self).get_parser(prog_name)
        parser.add_argument('execution', help='Execution ID.')

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.function_engine
        log = client.function_executions.get_log(parsed_args.execution)

        self.app.stdout.write(log or "\n")



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from qinlingclient.common import base


class Runtime(base.Resource):
    pass


class RuntimeManager(base.ManagerWithFind):
    resource_class = Runtime

    def list(self, **kwargs):
        return self._list("/v1/runtimes", response_key='runtimes')

    def create(self, image, name=None, description=None,
               is_public=True, trusted=True):
        data = {'image': image, 'is_public': is_public, 'trusted': trusted}
        if name:
            data.update({'name': name})
        if description:
            data.update({'description': description})

        return self._create('/v1/runtimes', data)

    def delete(self, id):
        self._delete('/v1/runtimes/%s' % id)

    def get(self, id):
        return self._get('/v1/runtimes/%s' % id)

    def get_pool(self, id):
        return self._get('/v1/runtimes/%s/pool' % id)



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from osc_lib.tests import utils as osc_tests_utils

from qinlingclient.common import exceptions
from qinlingclient.osc.v1 import base
from qinlingclient.osc.v1 import runtime
from qinlingclient.tests.unit.osc.v1 import fakes


class TestRuntime(fakes.TestQinlingClient):

    def setUp(self):
        super(TestRuntime, self).setUp()
        # Get a shortcut
        self.client = self.app.client_manager.function_engine

        self.columns = base.RUNTIME_COLUMNS
        self.data = []

        self._runtimes = fakes.FakeRuntime.create_runtimes(count=3)
        for r in self._runtimes:
            self.data.append((r.id, r.name, r.image, r.status, r.description,
                              r.is_public, r.trusted, r.project_id,
                              r.created_at, r.updated_at))


class TestListRuntime(TestRuntime):

    def setUp(self):
        super(TestListRuntime, self).setUp()
        self.cmd = runtime.List(self.app, None)

        self.columns = [c.capitalize() for c in base.RUNTIME_COLUMNS]

        self.client.runtimes.list = mock.Mock(return_value=self._runtimes)

    def test_runtime_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.client.runtimes.list.assert_called_once_with()
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_runtime_list_with_filter(self):
        arglist = ['--filter', 'name=has:runtime',
                   '--filter', 'status=eq:available']
        verifylist = [
            ('filters', ['name=has:runtime', 'status=eq:available'])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.client.runtimes.list.assert_called_once_with(
            name='has:runtime', status='eq:available'
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_runtime_list_with_invalid_filter(self):
        arglist = ['--filter', 'name']
        verifylist = [
            ('filters', ['name'])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaisesRegex(
            ValueError,
            '^Invalid filter: name$',
            self.cmd.take_action, parsed_args
        )


class TestCreateRuntime(TestRuntime):

    def setUp(self):
        super(TestCreateRuntime, self).setUp()
        self.cmd = runtime.Create(self.app, None)

    def _create_fake_runtime(self, attrs=None):
        # Allow to fake different create results
        r = fakes.FakeRuntime.create_one_runtime(attrs)
        self.client.runtimes.create = mock.Mock(return_value=r)
        data = (r.id, r.name, r.image, r.status, r.description,
                r.is_public, r.trusted, r.project_id,
                r.created_at, r.updated_at)
        return data

    def test_runtime_create_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_runtime_create_required_options(self):
        image = 'specified-image-name'
        attrs = {'image': image}
        created_data = self._create_fake_runtime(attrs)

        arglist = [image]
        verifylist = [
            ('image', image),
            ('name', None),
            ('description', None),
            ('trusted', True),
            ('is_public', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.runtimes.create.assert_called_once_with(**{
            'name': None,
            'description': None,
            'image': image,
            'trusted': True,
            'is_public': True,
        })
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)

    def test_runtime_create_all_options(self):
        image = 'specified-image-name'
        name = 'specified-runtime-name'
        description = 'specified-runtime-description'
        trusted = False
        is_public = False
        attrs = {'image': image, 'name': name,
                 'description': description, 'trusted': trusted,
                 'is_public': is_public}
        created_data = self._create_fake_runtime(attrs)

        arglist = [
            '--name', name,
            '--description', description,
            '--untrusted',
            '--private',
            image,
        ]
        verifylist = [
            ('image', image),
            ('name', name),
            ('description', description),
            ('trusted', trusted),
            ('is_public', is_public)
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.runtimes.create.assert_called_once_with(**{
            'name': name,
            'description': description,
            'image': image,
            'trusted': trusted,
            'is_public': is_public,
        })
        self.assertEqual(self.columns, columns)
        self.assertEqual(created_data, data)


class TestDeleteRuntime(TestRuntime):

    def setUp(self):
        super(TestDeleteRuntime, self).setUp()
        self.cmd = runtime.Delete(self.app, None)
        self.client.runtimes.delete = mock.Mock(return_value=None)

    def test_runtime_delete_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_runtime_delete_one(self):
        runtime_id = self._runtimes[0].id
        arglist = [runtime_id]
        verifylist = [('runtime', [runtime_id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        self.client.runtimes.delete.assert_called_once_with(runtime_id)

    def test_runtime_delete_multiple(self):
        runtime_ids = [r.id for r in self._runtimes]
        arglist = runtime_ids
        verifylist = [('runtime', runtime_ids)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)
        calls = [mock.call(r_id) for r_id in runtime_ids]
        self.assertEqual(len(runtime_ids),
                         self.client.runtimes.delete.call_count)
        self.client.runtimes.delete.assert_has_calls(calls)

    def test_runtime_delete_multiple_exception(self):
        runtime_ids = [r.id for r in self._runtimes]
        arglist = runtime_ids
        verifylist = [('runtime', runtime_ids)]

        self.client.runtimes.delete = mock.Mock(side_effect=[
            None, RuntimeError, None
        ])

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaisesRegex(
            exceptions.QinlingClientException,
            r'^Unable to delete the specified runtime\(s\)\.$',
            self.cmd.take_action, parsed_args)

        # The second deletion failed, but the third is done normally
        calls = [mock.call(r_id) for r_id in runtime_ids]
        self.assertEqual(len(runtime_ids),
                         self.client.runtimes.delete.call_count)
        self.client.runtimes.delete.assert_has_calls(calls)


class TestShowRuntime(TestRuntime):

    def setUp(self):
        super(TestShowRuntime, self).setUp()
        self.cmd = runtime.Show(self.app, None)
        self.client.runtimes.get = mock.Mock(return_value=self._runtimes[0])

    def test_runtime_show_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_runtime_show(self):
        runtime_id = self._runtimes[0].id
        arglist = [runtime_id]
        verifylist = [('runtime', runtime_id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.runtimes.get.assert_called_once_with(runtime_id)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data[0], data)


class TestShowRuntimePool(TestRuntime):

    def setUp(self):
        super(TestShowRuntimePool, self).setUp()
        self.cmd = runtime.Pool(self.app, None)

        self.columns = base.RUNTIME_POOL_COLUMNS
        pool_attrs = {'name': self._runtimes[0].id}
        pool = fakes.FakeRuntime.create_one_runtime_pool(pool_attrs)
        self.pool_data = (pool.name, pool.capacity)

        self.client.runtimes.get_pool = mock.Mock(return_value=pool)

    def test_runtime_pool_show_no_option(self):
        arglist = []
        verifylist = []

        self.assertRaises(osc_tests_utils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_runtime_pool_show(self):
        runtime_id = self._runtimes[0].id
        arglist = [runtime_id]
        verifylist = [('runtime', runtime_id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.client.runtimes.get_pool.assert_called_once_with(runtime_id)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.pool_data, data)



# Copyright 2018 AWCloud Software Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from keystoneauth1 import session
from requests_mock.contrib import fixture

from osc_lib.tests import utils

from qinlingclient.v1 import client

QINLING_URL = 'http://example.com:7070'


class TestQinlingClient(utils.TestCase):

    def setUp(self):
        super(TestQinlingClient, self).setUp()
        sess = session.Session()
        self.client = client.Client(QINLING_URL, session=sess)
        self.requests_mock = self.useFixture(fixture.Fixture())
